/*
    $Id: dump.c,v 1.6 2001/11/01 20:55:50 sensei Exp $

    Copyright (C) 2000-2001  Marko Zivanovic
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/
     
#ifndef lint
static char vcid[] __attribute__ ((__unused__)) = "$Id: dump.c,v 1.6 2001/11/01 20:55:50 sensei Exp $";
#endif /* lint */

#include "dump.h"

/*  ************************************************************ */
/*  dumper thread main controling function                       */
/*                                                               */
/*  parameters:                                                  */
/*                                                               */
/*     param   -   parameter specified during thread creation    */
/*                 (currently unused)                            */
/*                                                               */
/*  return:                                                      */
/*     void*   -   return value; picked up with pthread_join     */
/*  ************************************************************ */
void *start_dumper (void *param)
{
  time_t last_dump, now;
  long pkt_count;

  while (stop_dumper == 0) {
    /* be extremely nice while idle */
    setpriority (PRIO_PROCESS, 0, 20);
    time (&last_dump);
    pkt_count = no_of_packets;
    while (1) {
      if (stop_dumper == 1) break;
      sleep (1);

      if (immediate_dump == 1) {
	immediate_dump = 0;
	break;
      }

      /* check if more than dump_freq seconds have elapsed since the
         last dump and if so, dump the buffer */
      time (&now);
      if (abs (difftime (now, last_dump)) >= dump_freq) {
	break;
      }
      
      /* check if buffer has grown beyond dump_buffer_size limit; if
         it has, dump the buffer */
      if (buffer_length >= dump_buffer_size * 1024) {
	break;
      }

      /* check if number of buffered packets has exceeded dump_packets
         count and if it has, dump buffer */
      if (no_of_packets - pkt_count >= dump_packets) {
	break;
      }
    }
    
    setpriority (PRIO_PROCESS, 0, dumper_priority);
    dump_buffer ();
  }
  
  /* call dump_buffer once again just in case there are some packets
     left in the buffer */
  dump_buffer ();    

  pthread_exit (NULL);
}

/*  ************************************************************ */
/*  dump current contents of buffer                              */
/*                                                               */
/*  no parameters                                                */
/*  return:                                                      */
/*     int   -   unused for now                                  */
/*  ************************************************************ */
int dump_buffer ()
{
  eth_ip_frame *eth_ip;
  chain_buffer *cb_curr;
  
  cb_curr = cb_first;
  
  while (1) {
    if (cb_last == NULL)
      break; /* buffer is empty => exit */

    if ((stop_dumper == 0) && (cb_curr == cb_last))
      break; /* we're still running and the buffer is empty => exit */
    
    eth_ip = cb_curr->eth_ip_frame;
    
    /* call on_packet_dump function */
    S_on_dump_packet (eth_ip);

    /* if we should stop, dump the last remaining packet */
    if ((stop_dumper == 1) && (cb_first == cb_last)) {
      free (cb_curr->eth_ip_frame);
      free (cb_curr);
      cb_last = cb_first = cb_curr = NULL;
      break;
    }
    
    /*  remove the packet from the buffer */
    cb_first = cb_curr->next;
    buffer_length -= cb_curr->eth_ip_frame->length;
    no_of_packets_in_buffer--;
    free (cb_curr->eth_ip_frame);
    free (cb_curr);
    cb_curr = cb_first;
    
  } /* while cb_curr->next != cb_last */
  return 0;
}
