/*
    $Id: main.c,v 1.4 2001/11/01 20:55:50 sensei Exp $

    Copyright (C) 2000-2001  Marko Zivanovic
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef lint
static char vcid[] __attribute__ ((__unused__)) = "$Id: main.c,v 1.4 2001/11/01 20:55:50 sensei Exp $";
#endif /* lint */

#include "config.h"
#include "../config.h"
#include "sniffer.h"
#include "helper.h"
#include "dump.h"
#include "sighandlers.h"
#include "log.h"
#include "scripting.h"

#include <signal.h>
#include <pthread.h>
#include <stdio.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#else
#error "Cannot #include <unistd.h>"
#endif

#define _GNU_SOURCE
#include <getopt.h>

/* status variables */
time_t start_time, now;
long up, day, hour, min, sec;
long buffer_length = 0;
long no_of_packets = 0;
long total_buff_len = 0;
long no_of_packets_in_buffer = 0;
long total_packets = 0;

/* file hadles */
FILE *fout;
FILE *flog;

/* debug mode flags */
int debug_config_file = 0;
int debug_malloc = 0;
int debug_dont_fork = 0;

/* quiet mode flag */
int quiet = 0;

/* host name lookup flag */
int nodns = 1;

/* flags used to control working threads */
int stop_sniffer = 0;
int stop_dumper = 0;
int immediate_dump = 0;

/* thread controling structures */
pthread_t t_sniff;
pthread_t t_dump;

int main(int argc, char *argv[])
{
  int has_config = 0;
  char cfg_filename[1024];
  int t_ret, ret;
  void *t_retval;

  /*  command line options */
  int c;
  
  flog = stderr;

  while (1)
    {
      int option_index = 0;
      static struct option long_options[] = {
	{"help", 0, 0, 'h'},
	{"version", 0, 0, 'v'},
	{"debug", 1, 0, 'd'},
	{"quiet", 0, 0, 'q'},
	{"no-host-name-lookup", 0, 0, 'n'},
	{0, 0, 0, 0}
      };
      
      c = getopt_long (argc, argv, "-hvd:qn", long_options, &option_index);
      if (c == -1) break; 
      
      switch (c) {
      case 'h':
	show_help ();
	break; 
	
      case 'v':
	show_version ();
	break; 

      case 'd':
	switch (optarg[0]) {
	case 'c': 
	  debug_config_file = 1;
	  break;
	case 'm':
	  debug_malloc = 1;
	  enable_malloc_debug (0);
	  break;
	case 'b':
	  debug_dont_fork = 1;
	  break;
	}
	break;

      case 1:
	has_config = 1;
	strcpy (cfg_filename, optarg);
	break;
	
      case 'q':
	fclose (stderr);
	fclose (stdout);
	stderr = (FILE*) fopen ("/dev/null", "w");
	stdout = (FILE*) fopen ("/dev/null", "w");
	break;

      case 'n':
	nodns = 0;
	break;

      case '?':
	break; 
	
      default:
	fprintf (stderr, "Unrecognized option: %c", c);
	exit (0);
      }
    } 

  if (geteuid () != 0) {
    logmsg (LOG_CRIT, "only root can run %s", PACKAGE);
    exit (1);
  }

  if (!has_config) {
    fprintf (stderr, "script file not specified!\n");
    exit (1);
  } else {
    S_load_script (cfg_filename);
  }
    
  S_on_initialize ();
  
  if (debug_dont_fork == 0) {
    ret = fork ();
    if (ret == -1) {
      logmsg (LOG_CRIT, "fork error: %s", strerror (errno));
      exit (1);
    }
    if (ret > 0) { /* parent */
      setpgid (ret, ret);
      exit (1);
    }
  }

  /* child process */
  setpgid (0, 0);
  open_log ();
  
  logmsg (LOG_INFO, "sniffer starting up");
  time (&start_time);

  /* initialize threads */
  t_ret = pthread_create (&t_sniff, NULL, start_sniffer, (void *) "");
  if (t_ret != 0) {
    logmsg (LOG_CRIT, "error creating sniffer thread %d\n", t_ret);
    exit (1);
  }
  
  t_ret = pthread_create (&t_dump, NULL, start_dumper, (void*) "");
  if (t_ret != 0) {
    logmsg (LOG_CRIT, "error creating dumper thread %d\n", t_ret);
    exit (1);
  }
  
  signal (SIGHUP, reload_configuration);
  signal (SIGUSR1, enable_malloc_debug);
  signal (SIGUSR2, disable_malloc_debug);
  signal (SIGINT, print_status);
  signal (SIGTERM, shutdown_sniffer);
  signal (SIGQUIT, shutdown_sniffer);
  signal (SIGWINCH, dump_buffer_now);

  t_ret = pthread_join (t_sniff, &t_retval);
  if (t_ret != 0) {
    logmsg (LOG_CRIT, "error joining sniffer thread %d\n", t_ret);
    exit (1);
  }

  t_ret = pthread_join (t_dump, &t_retval);
  if (t_ret != 0) {
    logmsg (LOG_CRIT, "error joining dumper thread %d\n", t_ret);
    exit (1);
  }

  S_on_shutdown ();

  time (&now);
  up = difftime (now, start_time);
  sec = up;
  day = sec / 86400;
  sec -= day * 86400;
  hour = sec / 3600;
  sec -= hour * 3600;
  min = sec / 60;
  sec -= min * 60;

  logmsg (LOG_INFO, "sniffer is down");
  logmsg (LOG_INFO, "session duration: %ld days, %ld hours, %ld minutes and %ld seconds", day, hour, min, sec);
  logmsg (LOG_INFO, "packets dumped: %ld", no_of_packets);
  logmsg (LOG_INFO, "total packets processed: %ld", total_packets);
  logmsg (LOG_INFO, "average load: %ld packets/s, %ld KB/s", no_of_packets / up, total_buff_len / up / 1024);

  return 0;
}
